import java.io.*;
import java.util.*;

/**
 * MannWhitneyU - a Java utility to perform a Mann-Whitney U test on a table of data
 * <p>
 * 
 * The Mann-Whitney U test is a non-parametric statistical test used to assess the difference
 * between two independent samples, which is to say, between two conditions assigned
 * between-subjects. The relationship with other non-parametric tests is illustrated below:
 * <p>
 * 
 * <blockquote><img src="MannWhitneyU-0.jpg"></blockquote>
 * <p>
 * 
 * The test operates on ordinal data, such as rankings or questionnaire responses (where the
 * responses are non-linear).
 * 
 * Interval or ratio-scale data can be used, however the Mann-Whitney U test (and other
 * non-parametric tests) pre-process the data into ranks. If the researcher believes the ANOVA
 * assumptions of normality and homogeneity of variances are violated to the extent that the ANOVA
 * is inappropriate as a test instrument, then a non-parametric test is a reasonable alternative.
 * However, as Sheskin notes, "when a researcher elects to transform a set of interval/ratio data
 * into ranks, information is sacrificed" (Sheskin, 1996, p. 423). A simple and obvious example is
 * the transformation of student marks A = 49, B = 79, and C = 80, into ranks 1, 2, and 3. Yes,
 * something is sacrificed. Others have noted that the parametric ANOVA is robust and relatively
 * unperturbed by violations in its assumptions (e.g., Lowry, Subchapter 14a in
 * "Concepts and Applications of Inferential Statistics";
 * http://vassarstats.net/textbook/index.html). So, the choice of a non-parametric test, such as the
 * Mann-Whitney U, should not be taken lightly if the data are ratio scale or interval scale.
 * <p>
 * 
 * Invocation (usage message if invoked without arguments):
 * <p>
 * 
 * <blockquote><a href="MannWhitneyU-2.jpg"><img src="MannWhitneyU-2.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * <b>Example.</b> Consider an experiment seeking to determine if there is a difference in the
 * political leaning of MAC users vs. PC users. Are MAC or PC users more likely to have political
 * views leaning to the "left" or "right"? The experimenter randomly selects 10 MAC users and 10 PC
 * users. The participants are interviewed and asked a variety of questions about their political
 * leaning. Each participant is assessed on a 10-point linear scale (1 = very left, 10 = very right)
 * with the results stored in a text file as a matrix with two columns. The first column contains
 * the assessments for the MAC users, the second column for the PC users (see <a
 * href="mannwhitneyu-ex1.txt">mannwhitneyu-ex1.txt</a>). The data are then processed by the
 * <code>MannWhitneyU</code> utility. For example,
 * <p>
 * 
 * <blockquote><a href="MannWhitneyU-3.jpg"><img src="MannWhitneyU-3.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * For convenience, the test statistic U is converted to a <i>z</i>-score. This facilitates
 * computing <i>p</i>. Since the customary threshold for statistical significance (<i>p</i> < .05)
 * was not exceeded, the null hypothesis remains tenable, inferring a likelihood that the two
 * samples are from the same population. We conclude there is no difference between the political
 * leaning of MAC users and PC users.
 * <p>
 * 
 * The test also provides <i>z'</i> which is corrected for ties. The difference is usually minor.
 * <p>
 * 
 *  StatView provides the following output using the data above:
 * <p>
 * 
 * <blockquote><img src="MannWhitneyU-1.jpg"></blockquote>
 * <p>
 * 
 * <b>Missing data.</b> Since the design is between-subjects, it is sometimes difficult to have
 * equal-size groups. For example, if two of the MAC participants do not show up for the interview,
 * then there will be 8 participants in the MAC group and 10 participants in the PC group. This is
 * accommodated by using periods (".") for missing data (see <a
 * href="mannwhitneyu-ex2.txt">mannwhitneyu-ex2.txt</a>). For example,
 * 
 * <blockquote><a href="MannWhitneyU-4.jpg"><img src="MannWhitneyU-4.jpg"
 * width="600"></a></blockquote>
 * <p>
 * 
 * The conclusion is the same. Note that "equal sample sizes" is one assumption of the parametric
 * ANOVA that, if violated, justifies using a non-parametric test, such as the Mann-Whitney U (see
 * Lowry reference, cited above).
 * <p>
 * 
 * @author Scott MacKenzie, 2012-2014
 */
public class MannWhitneyU
{
	public static void main(String[] args) throws IOException
	{
		double u, z1, p1, z2, p2;
		boolean vOption = false;

		if (args.length < 1)
		{
			System.out.println("------------------------------------------------------------------------\n"
					+ "Usage: java MannWhitneyU file [-v]\n" + "\n"
					+ "  where \'file\' contains a data table (comma or space delimited)\n"
					+ "        [-v]   = verbose (debugging and other information)\n"
					+ "------------------------------------------------------------------------\n");
			System.exit(0);
		}
		String file = args[0];
		if (args.length == 2 && args[1].equals("-v"))
			vOption = true;

		BufferedReader br = null;
		try
		{
			br = new BufferedReader(new FileReader(file));
		} catch (FileNotFoundException e)
		{
			System.out.println("File not found: " + file);
			System.exit(0);
		}

		String s = null;
		Vector<String> v = new Vector<String>();
		double[][] data;
		StringTokenizer st;
		int numberOfColumns = -1;

		// read the data
		try
		{
			while ((s = br.readLine()) != null)
				v.add(s);
		} catch (IOException e)
		{
			System.out.println("IOException: " + e);
			System.exit(0);
		}

		// determine number of rows and colunns
		int nRows = v.size();
		String tmp = v.elementAt(0);
		int nCols = (new StringTokenizer(tmp)).countTokens();

		if (nCols != 2)
		{
			System.out.println("Data format error.  Two columns required.");
			System.exit(0);
		}

		// create data matrix
		data = new double[nRows][nCols];

		// fill data matrix
		for (int i = 0; i < data.length; ++i)
		{
			st = new StringTokenizer(v.elementAt(i));
			if (numberOfColumns == -1)
				numberOfColumns = st.countTokens();
			else if (numberOfColumns != 2)
			{
				System.out.println("Error: Data format error.  Columns = " + st.countTokens());
				System.exit(0);
			}

			int j = 0;
			while (st.hasMoreTokens())
			{
				String ss = st.nextToken();
				if (ss.equals("."))
					data[i][j++] = Double.NaN;
				else
					data[i][j++] = Double.parseDouble(ss);
			}
		}

		if (vOption)
		{
			System.out.println("Raw data...");
			for (int i = 0; i < data.length; ++i)
			{
				for (int j = 0; j < data[0].length; ++j)
					System.out.print(data[i][j] + ", ");
				System.out.println();
			} // data matrix loaded!
			System.out.println();
		}

		MannWhitneyU mwu = new MannWhitneyU(data, vOption);
		u = mwu.getU();
		z1 = mwu.getZ1();
		z2 = mwu.getZ2();
		p1 = mwu.getP1();
		p2 = mwu.getP2();

		System.out.printf("U = %f\n", u);
		System.out.printf("z = %.3f, p = %.4f\n", z1, p1);
		System.out.printf("z' = %.3f, p' = %.4f\n", z2, p2);
	}

	public MannWhitneyU(double[][] dataArg, boolean vOption)
	{
		double[][] data = new double[dataArg.length][dataArg[0].length];
		for (int i = 0; i < data.length; ++i)
			for (int j = 0; j < data[0].length; ++j)
				data[i][j] = dataArg[i][j];

		int nRows = data.length;
		int nCols = data[0].length;

		double meanColumn1 = 0.0;
		double meanColumn2 = 0.0;
		int n1 = 0;
		int n2 = 0;
		for (int i = 0; i < nRows; ++i)
		{
			if (!Double.isNaN(data[i][0]))
			{
				meanColumn1 += data[i][0];
				++n1;
			}
			if (!Double.isNaN(data[i][1]))
			{
				meanColumn2 += data[i][1];
				++n2;
			}
		}
		meanColumn1 /= n1;
		meanColumn2 /= n2;
		int n = n1 + n2;

		if (vOption)
		{
			System.out.println("column means = " + meanColumn1 + ", " + meanColumn2);
			System.out.println();
		}

		double[] tmp2 = new double[nRows * nCols];
		int countNaN = 0;
		for (int i = 0; i < tmp2.length; ++i)
		{
			tmp2[i] = data[i % nRows][i / nRows];
			if (Double.isNaN(tmp2[i]))
				++countNaN;
		}

		double[] tmp3 = new double[tmp2.length - countNaN];
		for (int i = 0, j = 0; i < tmp2.length; ++i)
			if (!Double.isNaN(tmp2[i]))
				tmp3[j++] = tmp2[i];

		double[] ranks = new double[nRows * nCols];
		ranks = getRanks(tmp3);

		double[] ranksColumn1 = new double[nRows];
		double[] ranksColumn2 = new double[nRows];

		for (int i = 0, j = 0; i < nRows * nCols; ++i)
		{
			if (i / nRows == 0) // find ranks for left column
			{
				if (!Double.isNaN(tmp2[i]))
					ranksColumn1[i] = ranks[j++];
				else
					ranksColumn1[i] = Double.NaN;
			} else
			// find ranks for right column
			{
				if (i / nRows == 1 && !Double.isNaN(tmp2[i]))
					ranksColumn2[i - nRows] = ranks[j++];
				else
					ranksColumn2[i - nRows] = Double.NaN;
			}
		}

		if (vOption)
		{
			System.out.println("Rank data...");
			for (int i = 0; i < nRows; ++i)
				System.out.println(ranksColumn1[i] + ", " + ranksColumn2[i]);
			System.out.println();
		}

		double t1 = 0.0;
		double t2 = 0.0;
		for (int i = 0; i < nRows; ++i)
		{
			t1 += Double.isNaN(ranksColumn1[i]) ? 0.0 : ranksColumn1[i];
			t2 += Double.isNaN(ranksColumn2[i]) ? 0.0 : ranksColumn2[i];
		}

		double u1 = n1 * n2 + (n1 * (n1 + 1) / 2) - t1;
		double u2 = n1 * n2 + (n2 * (n2 + 1) / 2) - t2;
		u = Math.min(u1, u2);

		if (vOption)
		{
			System.out.println("t1 = " + t1);
			System.out.println("t2 = " + t2);
			System.out.println("n1 = " + n1);
			System.out.println("n2 = " + n2);
			System.out.println("N = " + n + " (N = n1 + n2)");
			System.out.println("U1 = " + u1);
			System.out.println("U2 = " + u2);
			System.out.println();
		}

		z1 = (u - (n1 * n2) / 2.0) / Math.sqrt(n1 * n2 * (n1 + n2 + 1.0) / 12.0);
		p1 = 2.0 * Statistics.normalProbability(z1);

		Arrays.sort(ranks);
		double factor = (n1 * n2 * correctionForTies(ranks)) / (12.0 * (n1 + n2) * (n1 + n2 - 1.0));
		z2 = (u - (n1 * n2) / 2.0) / Math.sqrt(n1 * n2 * (n1 + n2 + 1.0) / 12.0 - factor);
		p2 = 2.0 * Statistics.normalProbability(z2);

		r1 = Math.abs(z1) / Math.sqrt(n);
		r2 = Math.abs(z2) / Math.sqrt(n);
		if (vOption)
		{
			System.out.printf("r' = %.3f (effect size: r' = abs(z') / sqrt(N))\n", r2);
			System.out.printf("r = %.3f (effect size: r = abs(z) / sqrt(N))\n", r1);
			System.out.println();
		}
	}

	private double u, z1, p1, r1, z2, p2, r2;

	public double getU()
	{
		return u;
	}

	public double getZ1()
	{
		return z1;
	}

	public double getZ2()
	{
		return z2;
	}

	public double getP1()
	{
		return p1;
	}

	public double getP2()
	{
		return p2;
	}

	// given an array of data, returns a array of ranks (handles ties)
	public double[] getRanks(double[] d)
	{
		// put raw ranks in 'r' array
		double[] r = new double[d.length];

		// create a copy of 'd' array
		double[] a = new double[d.length];
		for (int i = 0; i < a.length; ++i)
			a[i] = d[i];
		Arrays.sort(a); // sort, but leave values intact

		for (int i = 0; i < a.length; ++i)
		{
			double rawScore = a[i];
			int rawRank = i + 1; // raw rank
			int count = 1;
			int j = i + 1;
			while (j < a.length) // look for ties (potentially to end of array)
			{
				if (rawScore == a[j]) // found a tie
				{
					rawRank += j + 1;
					++count;
					++j;
				} else
					// no tie, skip to next position in array
					break;
			}
			double rank = (double)rawRank / count;
			r[i] = rank;
			while (--count > 0) // continue to fill slots with ties
			{
				++i;
				r[i] = rank;
			}
		}

		// use a (raw scores, sorted), r (ranks, sorted), and d (raw scores, unsorted) to build rr
		double[] rr = new double[d.length];

		for (int i = 0; i < d.length; ++i)
		{
			for (int j = 0; j < d.length; ++j)
			{
				if (d[i] == a[j])
				{
					rr[i] = r[j];
					break;
				}
			}
		}
		return rr;
	}

	private static double correctionForTies(double[] rArg)
	{
		double factor = 0.0;

		double[] r = new double[rArg.length];
		for (int i = 0; i < r.length; ++i)
			r[i] = rArg[i];
		Arrays.sort(r);

		for (int i = 0; i < r.length; ++i)
		{
			double rank = r[i];
			int j = i + 1;
			int countRanks = 1;
			while (j < r.length)
			{
				if (rank == r[j])
				{
					// System.out.println("FOUND TIE: i=" + i + ", j=" + j + ", k=" + k + ", rank="
					// + rank);
					++countRanks;
					++i;
					++j;
				} else
					break;
			}
			if (countRanks > 1)
				factor += countRanks * countRanks * countRanks - countRanks;
		}
		return factor;
	}
}
